import * as XLSX from "./";
import * as assert from "assert";
import numbers from "./dist/xlsx.zahl";
var JSDOM; try { JSDOM = require("jsdom").JSDOM; } catch(e) {}
var version = process.version;

describe("API Functions", function() {
  var basedate = new Date(Date.UTC(2048,9,6,0,0,0,0));
  var basedate_local = new Date(2048,9,6,0,0,0,0);
  var basedate_back = new Date(basedate.valueOf() * 2 - basedate_local.valueOf());
  function doit(utc: boolean) {
    describe("should handle " + (utc ? "UTC" : "local") + " dates", function() {
      var sharedate = utc ? basedate : basedate_local;
      var sharedate1 = utc ? basedate_back : basedate;
      it("aoa_to_sheet", function() {
        var data = [["Date"], [sharedate]];
        var ws = XLSX.utils.aoa_to_sheet(data, {UTC: utc}); assert.equal(ws["A2"].v, 54337);
        //var wb = XLSX.utils.book_new(); XLSX.utils.book_append_sheet(wb, ws, "Sheet1"); XLSX.writeFile(wb, "aoa" + (utc ? "true" : "false") + ".xlsx");
        ws = XLSX.utils.aoa_to_sheet(data, {cellDates: true, UTC: utc}); assert.equal(ws["A2"].v.valueOf(), basedate.valueOf());
        var js = XLSX.utils.sheet_to_json<any[]>(ws, {UTC: utc, header:1}); assert.equal(js[1][0].valueOf(), sharedate.valueOf());
        js = XLSX.utils.sheet_to_json(ws, {UTC: true, header:1}); assert.equal(js[1][0].valueOf(), basedate.valueOf());
        js = XLSX.utils.sheet_to_json(ws, {UTC: false, header:1}); assert.equal(js[1][0].valueOf(), basedate_local.valueOf());
      });
      interface IDate { Date: Date; };
      it("json_to_sheet", function() {
        var data = [{Date: sharedate}];
        var ws = XLSX.utils.json_to_sheet(data, {UTC: utc}); assert.equal(ws["A2"].v, 54337);
        ws = XLSX.utils.json_to_sheet(data, {cellDates: true, UTC: utc}); assert.equal(ws["A2"].v.valueOf(), basedate.valueOf());
        var js = XLSX.utils.sheet_to_json<IDate>(ws, {UTC: utc}); assert.equal(js[0].Date.valueOf(), sharedate.valueOf());
        js = XLSX.utils.sheet_to_json(ws, {UTC: true}); assert.equal(js[0].Date.valueOf(), basedate.valueOf());
        js = XLSX.utils.sheet_to_json(ws, {UTC: false}); assert.equal(js[0].Date.valueOf(), basedate_local.valueOf());
      });
      /* NOTE: local time not supported in sheet_to_html or HTML parser */
      it("HTML format", function() {
        var data = "<table><tr><td>Date</td></tr><tr><td>" + "2048-10-06" + "</td></tr></table>";
        var ws = XLSX.read(data, {UTC: utc, cellDates: true, type: "binary"}).Sheets.Sheet1; assert.equal(ws["A2"].v.valueOf(), sharedate.valueOf());
        var js = XLSX.utils.sheet_to_html(ws, {}); if(utc) assert(js.indexOf(sharedate.toISOString())>-1);
      });
      (JSDOM ? it : it.skip)("table_to_sheet", function() {
        var data = "<table><tr><td>Date</td></tr><tr><td>" + "2048-10-06" + "</td></tr></table>";
        var dom = new JSDOM(data); var tbl = dom.window.document.querySelector("table");
        //var wb = XLSX.utils.table_to_book(tbl, {UTC: true, cellDates: true}); XLSX.writeFile(wb, "dafuqtrue.xlsx");
        //var wb = XLSX.utils.table_to_book(tbl, {UTC: false, cellDates: true}); XLSX.writeFile(wb, "dafuqfalse.xlsx");
        var ws = XLSX.utils.table_to_sheet(tbl, {UTC: utc, cellDates: true}); assert.equal(ws["A2"].v.valueOf(), sharedate1.valueOf());
        var js = XLSX.utils.sheet_to_html(ws, {}); assert(js.indexOf(sharedate1.toISOString()) > -1);
      });
      it("CSV format", function() {
        var data = "Date\n2048-10-06";
        var ws = XLSX.read(data, {UTC: utc, cellDates: true, type: "binary"}).Sheets.Sheet1; assert.equal(ws["A2"].v.valueOf(), sharedate.valueOf());
        delete ws["A2"].w;
        var js = XLSX.utils.sheet_to_csv(ws, {}); assert.equal(js, "Date\n" + (!utc && new Date("2048-10-06").getTimezoneOffset() < 0 ? "10/5/48" : "10/6/48"));
        var ws2 = XLSX.read(js, {UTC: utc, cellDates: true, type: "binary"}).Sheets.Sheet1;
        if(!(sharedate.getTimezoneOffset() < 0 && !utc)) assert.equal(ws["A2"].v.valueOf(), ws2["A2"].v.valueOf());
      });
    });
  }
  doit(true);
  doit(false);
});
describe("SSF", function() {
  function doit(fmt: string|number, val: string, date1904 = false) {
    var basedate = date1904 ? new Date(Date.UTC(2052,9,7,0,0,0,0)) : new Date(Date.UTC(2048,9,6,0,0,0,0));
    var basetime = 54337;
    for(var i = 0; i < 23; ++i) {
      basedate.setUTCHours(i);
      for(var j = 0; j < 60; j+= 59) {
        basedate.setUTCMinutes(j);
        basedate.setUTCSeconds(j);
        assert.equal(XLSX.SSF.format(fmt, basedate), val);
        if(XLSX.SSF.is_date(fmt) || XLSX.SSF.is_date(XLSX.SSF._table[fmt]||"General")) assert.equal(XLSX.SSF.format(fmt, basetime + (i * 60 + j)/3600, {date1904: date1904}), val);
      }
    }
  }
  it("General", function() { doit("General", "10/6/48"); });
  it("14", function() { doit(14, "10/6/48"); });
  it("yyyy-mm-dd", function() { doit("yyyy-mm-dd", "2048-10-06"); })
  it("General 1904", function() { doit("General", "10/7/52", true); });
  it("14 1904", function() { doit(14, "10/7/52", true); });
  it("yyyy-mm-dd 1904", function() { doit("yyyy-mm-dd", "2052-10-07", true); })
});

var baseutc = [
  Date.UTC(2048,9,6,0,0,0,0),
  Date.UTC(1900,0,0,15,0,0,0),
  Date.UTC(1904,0,1,15,0,0,0),
];
var baselocal = [
  new Date(2048,9,6,0,0,0,0).valueOf(),
  new Date(1900,0,0,15,0,0,0).valueOf(),
  new Date(1904,0,1,15,0,0,0).valueOf(),
];
function check_ws_dates(ws: XLSX.WorkSheet, dbf = false, qpw = false, local = false, t1904 = false) {
  var base = local ? baselocal : baseutc;
  assert.equal(ws["B" + (dbf ? "2" : "1")].t, "d");
  assert.equal(ws["B" + (dbf ? "2" : "1")].v.valueOf(), base[0]);
  assert.equal(ws["B" + (dbf ? "3" : "2")].t, "d");
  if(t1904) assert.equal(ws["B2"].v.valueOf(), base[2]);
  else if(!dbf) assert.equal(ws["B2"].v.valueOf(), base[1]);
  else assert.equal(ws["B3"].v.valueOf(), Date.UTC(1900,0,0,0,0,0,0))
  if(ws[dbf ? "A4" : "A3"]) {
    if(qpw) {
      assert.equal(ws["B3"].t, "n");
      assert.equal(ws["B3"].v, 54337.625);
    } else {
      assert.equal(ws[dbf ? "B4" : "B3"].t, "d");
      assert.equal(ws[dbf ? "B4" : "B3"].v.valueOf(), local ? new Date(2048,9,6,dbf ? 0 : 15,0,0,0).valueOf() : Date.UTC(2048,9,6,dbf ? 0 : 15,0,0,0));
    }
  }
}

var b1904utc = [
  Date.UTC(2052,9,7,0,0,0,0),
  Date.UTC(1904,0,1,15,0,0,0),
  Date.UTC(1900,0,0,15,0,0,0),
];
var b1904local = [
  new Date(2052,9,7,0,0,0,0).valueOf(),
  new Date(1904,0,1,15,0,0,0).valueOf(),
  new Date(1900,0,0,15,0,0,0).valueOf(),
];
function check_ws_dates1904(ws: XLSX.WorkSheet, dbf = false, local = false, t1900 = false) {
  var base = local ? b1904local : b1904utc;
  assert.equal(ws["B" + (dbf ? "2" : "1")].t, "d");
  assert.equal(ws["B" + (dbf ? "2" : "1")].v.valueOf(), base[0]);
  assert.equal(ws["B" + (dbf ? "3" : "2")].t, "d");
  if(t1900) assert.equal(ws["B2"].v.valueOf(), base[2]);
  else if(!dbf) assert.equal(ws["B2"].v.valueOf(), base[1]);
  else assert.equal(ws["B3"].v.valueOf(), Date.UTC(1900,0,0,0,0,0,0))
  if(ws["A3"]) {
    assert.equal(ws[dbf ? "B4" : "B3"].t, "d");
    assert.equal(ws[dbf ? "B4" : "B3"].v.valueOf(), local ? new Date(2052,9,7,dbf ? 0 : 15,0,0,0).valueOf() : Date.UTC(2052,9,7,dbf ? 0 : 15,0,0,0));
  }
}

function check_ws_codes(ws: XLSX.WorkSheet, dbf = false, qpw = false, local = false) {
  var delta1 = local ? (baselocal[0] - baseutc[0])/1000/60/60/24 : 0;
  var delta2 = local ? (new Date(1899,11,31,0,0,0,0).valueOf() - Date.UTC(1899,11,31,0,0,0,0))/1000/60/60/24 : 0;
  assert.equal(ws[dbf ? "B2" : "B1"].t, "n");
  assert.equal(Math.round(ws[dbf ? "B2" : "B1"].v*1000), Math.round(1000*(54337 + delta1)));
  assert.equal(ws[dbf ? "B3" : "B2"].t, "n");
  assert.equal(Math.round(ws[dbf ? "B3" : "B2"].v*1000), Math.round(1000 * ((dbf ? 0 : 5/8) + delta2)));
  if(ws[dbf ? "A4" : "A3"]) {
    assert.equal(ws[dbf ? "B4" : "B3"].t, "n");
    assert.equal(Math.round(1000 * ws[dbf ? "B4" : "B3"].v), Math.round(1000 * (54337 + (dbf ? 0 : 5/8) + delta1)));
  }
}
function check_ws_codes1904(ws: XLSX.WorkSheet, pass = false, local = false, t1900 = false) {
  var delta1 = local ? (b1904local[0] - b1904utc[0])/1000/60/60/24 : 0;
  var delta2 = local ? (new Date(1899,11,31,0,0,0,0).valueOf() - Date.UTC(1899,11,31,0,0,0,0))/1000/60/60/24 : 0;
  pass = pass || false;
  var dbf = false;
  assert.equal(ws[dbf ? "B2" : "B1"].t, "n");
  assert.equal(Math.round(ws[dbf ? "B2" : "B1"].v*1000), Math.round(1000*(54337 + (+pass)*1462 + delta1)));
  //assert.equal(ws["B1"].v, 54337 + pass*1462);
  assert.equal(ws[dbf ? "B3" : "B2"].t, "n");
  assert(
    Math.abs(ws["B2"].v - delta2 - 5/8) < 0.01
    || Math.abs(ws["B2"].v - delta2 - (1462 + 5/8)) < 0.01
    || Math.abs(ws["B2"].v - delta2 - (-1462 + 5/8)) < 0.01
  ); // ODS :(
  if(ws["A3"]) {
    assert.equal(ws["B3"].t, "n");
    assert.equal(ws["B3"].v, 54337 + 5/8 + delta1 + (+pass)*1462);
  }
}

describe("read", function() {
([
  ["isocsv", "date/iso.csv", 1],
  ["xlsx", "date/DateTime.xlsx"],
  ["isox", "date/DateTime.iso.xlsx"],
  ["xlsb", "date/DateTime.xlsb"],
  ["xls8", "date/DateTime.xls"],
  ["xls5", "date/DateTime.biff5.xls"],
  ["sylk", "date/DateTime.slk"],
  ["xlml", "date/DateTime.xml"],
  ["ods", "date/DateTime.ods"],
  ["csv", "date/DateTime.csv", 1],
  ["txt", "date/DateTime.txt", 1],
  ["htm", "date/DateTime.htm", 1],

  ["wks-works3", "date/works3/DATETIME.WKS"],
  ["wk1-works3", "date/works3/DATETIME.WK1"],
  ["wk1-lotus22", "date/lotus22/DATETIME.WK1"],

  ["qp4-wq1", "date/qp4/qp4.wq1"],
  // ["QP5-wq2", "date/QP5/QP5.WQ2"], // WQ2 has its own format table

  ["xl5-slk", "date/xl5/dt5.slk"],
  ["xl5-biff5", "date/xl5/dt5.xls"],
  ["xl5-biff4", "date/xl5/dt4.xls"],
  ["xl5-biff4w", "date/xl5/dt4.xlw"], // corrupted in Excel? why?
  ["xl5-biff3", "date/xl5/dt3.xls"],

  ["xl5-biff2", "date/xl5/dt2.xls"],
  ["xl5-dif", "date/xl5/dt5.dif"],
  ["xl5-dbf4", "date/xl5/dt4.dbf"],
  ["xl5-dbf3", "date/xl5/dt3.dbf"],
  //["xl5-dbf2", "date/xl5/dt2.dbf", 0], // wrong data -- see file
  //["xl5-wq1", "date/xl5/dt5.wq1"], // wrong format for B2/B4 -- check file in QP
  //["xl5-wks", "date/xl5/dt5.wks"],
  //["xl5-wk1", "date/xl5/dt5.wk1"], // WK1 did not write correct Date format codes
  //["xl5-wk3", "date/xl5/dt5.wk3"],
  ["wps-xlsx", "date/wps/wps.xlsx"],
  ["wps-xls", "date/wps/wps.xls"],
  ["wps-et", "date/wps/wps.et"],
  ["wps-xlml", "date/wps/wps.xml"],
  ["wps-csv", "date/wps/wps.csv", 1],

  //["works9-csv", "date/works9/works9.csv", 0], // wrong data -- see file
  //["works9-wks", "date/works9/works9.wks"], // ??
  ["works9-xlsb", "date/works9/works9.xlsb"],
  ["works9-xlr", "date/works9/works9.xlr"],

  //["QP11-slk", "date/QP11/QP11.slk"], // missing formats
  //["QP11-wq1", "date/QP11/QP11.wq1"], // wrong format for B3
  //["QP11-wb1", "date/QP11/QP11.wb1"], // ??
  //["QP11-wb2", "date/QP11/QP11V6.wb2"], // ??
  //["QP11-wb3", "date/QP11/QP11V7.wb3"], // ??
  ["QP11-qpw", "date/QP11/QP11V9.qpw"], // ??

  ["123-biff4", "date/R9/r4.xls"],
  ["123-biff5", "date/R9/r5.xls"],
  ["123-biff8", "date/R9/r97.xls"],
  ["123-wk1", "date/R9/r9.wk1"],
  //["123-wk3", "date/R9/r9.wk3"], // ??
  //["123-wk4", "date/R9/r9.wk4"], // ??
  //["123-warp", "date/R9/Rwarp.123"], // ??
  //["123-97", "date/R9/R97.123"], // ??
  //["123-9", "date/R9/R9.123"], // ??
  // rtf
] as Array<[string, string] | [string, string, any]>).forEach(function(row) {
  describe(row[0], function() {
    it('should read UTC objects', function() {
      var wb = XLSX.readFile(row[1], {cellDates: true, WTF: true, UTC: true});
      //if(row[0] == "xl5-biff4w") console.log(wb, wb.Sheets[wb.SheetNames[[0]]])
      check_ws_dates(wb.Sheets[wb.SheetNames[0]], row[1].slice(-4) == ".dbf", row[1].slice(-4) == ".qpw", false);
    });
    it('should read UTC codes', function() {
      var wb = XLSX.readFile(row[1], {cellDates: false, WTF: true, UTC: true});
      //if(row[0] == "xl5-biff4w") console.log(wb.Sheets[wb.SheetNames[[0]]])
      check_ws_codes(wb.Sheets[wb.SheetNames[0]], row[1].slice(-4) == ".dbf");
    });
    it('should read date objects', function() {
      var wb = XLSX.readFile(row[1], {cellDates: true, WTF: true});
      //if(row[0] == "xl5-biff4w") console.log(wb, wb.Sheets[wb.SheetNames[[0]]])
      check_ws_dates(wb.Sheets[wb.SheetNames[0]], row[1].slice(-4) == ".dbf", row[1].slice(-4) == ".qpw", false);
    });
    it('should read date codes', function() {
      var wb = XLSX.readFile(row[1], {cellDates: false, WTF: true});
      //if(row[0] == "xl5-biff4w") console.log(wb.Sheets[wb.SheetNames[[0]]])
      check_ws_codes(wb.Sheets[wb.SheetNames[0]], row[1].slice(-4) == ".dbf", row[1].slice(-4) == ".qpw", false);
    });
    if(row[2] == 1) {
      it('should read local objects', function() {
        var wb = XLSX.readFile(row[1], {cellDates: true, WTF: true, UTC: false});
        //if(row[0] == "xl5-biff4w") console.log(wb, wb.Sheets[wb.SheetNames[[0]]])
        check_ws_dates(wb.Sheets[wb.SheetNames[0]], row[1].slice(-4) == ".dbf", row[1].slice(-4) == ".qpw", true);
      });
      it('should read local codes', function() {
        var wb = XLSX.readFile(row[1], {cellDates: false, WTF: true, UTC: false});
        //if(row[0] == "xl5-biff4w") console.log(wb, wb.Sheets[wb.SheetNames[[0]]])
        check_ws_codes(wb.Sheets[wb.SheetNames[0]], row[1].slice(-4) == ".dbf", row[1].slice(-4) == ".qpw", true);
      });
    }
  });
}); });

describe("read1904", function() {
  ([
    ["numbers", "date1904/DateTime.numbers"],
    ["xlsx", "date1904/DateTime.xlsx"],
    ["isox", "date1904/DateTime.iso.xlsx"],
    ["xlsb", "date1904/DateTime.xlsb"],
    ["xls8", "date1904/DateTime.xls"],
    ["xls5", "date1904/DateTime.biff5.xls"],
    ["sylk", "date1904/DateTime.slk"],
    ["xlml", "date1904/DateTime.xml"],
    ["ods", "date1904/DateTime.ods"],
    ["csv", "date1904/DateTime.csv", 1, 1],
    ["txt", "date1904/DateTime.txt", 1, 1],
    ["htm", "date1904/DateTime.htm", 1, 1],

    ["xl5-slk", "date1904/xl5/dt5.slk"],
    ["xl5-biff5", "date1904/xl5/dt5.xls"],
    ["xl5-biff4", "date1904/xl5/dt4.xls"],
    ["xl5-biff4w", "date1904/xl5/dt4.xlw"], // corrupted
    ["xl5-biff3", "date1904/xl5/dt3.xls"],
    ["xl5-biff2", "date1904/xl5/dt2.xls"],
  ] as Array<[string, string] | [string, string, number, number]>).forEach(function(row) {
    if(row[0] == "numbers" && version.indexOf("v0") > -1) return;
    describe(row[0], function() {
      if(!row[2]) it('should have date1904 set', function() {
        var wb = XLSX.readFile(row[1], {cellDates: true, WTF: true, UTC: true});
        assert(wb?.Workbook?.WBProps?.date1904);
      });
      it('should read UTC date objects', function() {
        var wb = XLSX.readFile(row[1], {cellDates: true, WTF: true, UTC: true});
        //if(row[0] == "xl5-biff4w") console.log(wb, wb.Sheets[wb.SheetNames[[0]]])
        check_ws_dates1904(wb.Sheets[wb.SheetNames[0]], false, false, !!row[3]);
      });
      it('should read UTC date codes', function() {
        var wb = XLSX.readFile(row[1], {cellDates: false, WTF: true, UTC: true});
        //if(row[0] == "xl5-biff4w") console.log(wb.Sheets[wb.SheetNames[[0]]])
        check_ws_codes1904(wb.Sheets[wb.SheetNames[0]], !!row[2], false, !!row[3]);
      });
      if(row[2]) it('should read local date objects', function() {
        var wb = XLSX.readFile(row[1], {cellDates: true, WTF: true, UTC: false});
        //if(row[0] == "xl5-biff4w") console.log(wb, wb.Sheets[wb.SheetNames[[0]]])
        check_ws_dates1904(wb.Sheets[wb.SheetNames[0]], false, true, !!row[3]);
      });
      if(row[2]) it('should read local date codes', function() {
        var wb = XLSX.readFile(row[1], {cellDates: false, WTF: true, UTC: false});
        //if(row[0] == "xl5-biff4w") console.log(wb.Sheets[wb.SheetNames[[0]]])
        check_ws_codes1904(wb.Sheets[wb.SheetNames[0]], !!row[2], true, !!row[3]);
      });
    });
  });
});

describe("roundtrip", function() {
  ([
    ["xlsx", "date/DateTime.xlsx", "xlsx"],
    ["xlsb", "date/DateTime.xlsb", "xlsb"],
    ["xls8", "date/DateTime.xls", "xls"],

    ["sylk", "date/DateTime.slk", "slk"], // TODO: write number format
    ["xlml", "date/DateTime.xml", "xlml"],
    ["ods", "date/DateTime.ods", "ods"],
    ["csv", "date/DateTime.csv", "csv"],
    ["txt", "date/DateTime.txt", "txt"],
    ["htm", "date/DateTime.htm", "htm"],

    ["xls5", "date/xl5/dt5.xls", "biff5"],
    ["xls4", "date/xl5/dt4.xls", "biff4"],
    ["xls3", "date/xl5/dt3.xls", "biff3"],
    ["xls2", "date/xl5/dt2.xls", "biff2"],
    ["dbf", "date/xl5/dt4.dbf", "dbf"],
  ] as Array<[string, string, XLSX.BookType]>).forEach(function(row) {
    describe(row[0], function() {
      [0,1,2,3].forEach(function(i) {it(String(i), function() {
        var wb = XLSX.readFile(row[1], {cellDates: !!(i & 1), WTF: true, cellNF: true, UTC: true});
        //if(row[0] == "sylk") console.log(wb.Sheets);
        var buf = XLSX.write(wb, { bookType: row[2], cellDates: !!(i&2), type: "buffer", WTF: true, numbers: numbers });
        //require("fs").writeFileSync(row[1] + "." + i + "." + row[2], buf);
        var wb3 = XLSX.read(buf, { cellDates: false, WTF: true, UTC: true});
        //if(row[0] == "sylk") console.log(wb3.Sheets);
        check_ws_codes(wb3.Sheets[wb3.SheetNames[0]], row[2] == "dbf");
        if(buf.l) delete buf.l;
        var wb2 = XLSX.read(buf, { cellDates: true, WTF: true, UTC: true});
        //if(row[0] == "sylk") console.log(wb2.Sheets);
        check_ws_dates(wb2.Sheets[wb2.SheetNames[0]], row[2] == "dbf");

        // TODO: UTC false

      });});
    });
  });
});

/*
    ["csv", "date1904/DateTime.csv", 1],
    ["txt", "date1904/DateTime.txt", 1],
    ["htm", "date1904/DateTime.htm", 1],
*/
describe("roundtrip1904", function() {
  ([
    ["n2xlsx", "date1904/DateTime.numbers", "xlsx"],
    ["xlsx", "date1904/DateTime.xlsx", "xlsx"],
    ["xlsb", "date1904/DateTime.xlsb", "xlsb"],
    ["xls8", "date1904/DateTime.xls", "xls"],
    ["xls5", "date1904/DateTime.biff5.xls", "biff5"],

    ["sylk", "date1904/DateTime.slk", "slk"], // TODO: write number format
    ["xlml", "date1904/DateTime.xml", "xlml"],
    ["ods", "date1904/DateTime.patched.ods", "fods", 1],
    /*["csv", "date/DateTime.csv", "csv"],
    ["txt", "date/DateTime.txt", "txt"],
    ["htm", "date/DateTime.htm", "htm"],*/

    ["xls5", "date1904/xl5/dt5.xls", "biff5"],
    ["xls4", "date1904/xl5/dt4.xls", "biff4"],
    ["xls3", "date1904/xl5/dt3.xls", "biff3"],
    ["xls2", "date1904/xl5/dt2.xls", "biff2"],

    ["numbers", "date1904/DateTime.numbers", "numbers"],
    ["xlsx2n", "date1904/DateTime.xlsx", "numbers"],

    //["wk1-works3", "date/works3/DATETIME.WK1", "wk1"],
  ] as Array<[string, string, XLSX.BookType]|[string, string, XLSX.BookType, number]>).forEach(function(row) {
    if((row[0] == "numbers" || row[1].indexOf("numbers") > -1 || row[2] == "numbers")&& version.indexOf("v0") > -1) return;
    describe(row[0], function() {
      [0,1,2,3].forEach(function(i) {it(String(i), function() {
        var wb = XLSX.readFile(row[1], {cellDates: !!(i & 1), WTF: true, cellNF: true});
        assert(wb?.Workbook?.WBProps?.date1904);
        var buf = XLSX.write(wb, { bookType: row[2], cellDates: !!(i&2), type: "buffer", WTF: true, numbers: numbers });
        //require("fs").writeFileSync(row[1] + "." + i + "." + row[2], buf);
        //console.log(wb.Workbook.WBProps.date1904, wb.Sheets.Sheet1.B1);

        if(buf.l) delete buf.l;
        var wb3 = XLSX.read(buf, { cellDates: false, cellNF: true, WTF: true});
        assert(wb3?.Workbook?.WBProps?.date1904);
        //console.log(wb3.Workbook.WBProps.date1904, wb3.Sheets.Sheet1.B1);
        check_ws_codes1904(wb3.Sheets[wb3.SheetNames[0]]);

        if(buf.l) delete buf.l;
        var wb2 = XLSX.read(buf, { cellDates: true, WTF: true});
        assert(wb2.Workbook?.WBProps?.date1904);
        //console.log(wb2.Workbook.WBProps.date1904, wb2.Sheets.Sheet1.B1);
        check_ws_dates1904(wb2.Sheets[wb2.SheetNames[0]], false, false, !!row[3] && !!(i&1));
      });});
    });
  });
});

describe("cross1904", function() {
  it("should cross 1900 -> 1904", function() {
  (["xlsx", "xlsb", "xls", "xlml", "biff5", "ods", "slk"] as Array<XLSX.BookType>).forEach(function(bookType) {
    var wb = XLSX.readFile("date/DateTime.xlsx", {cellDates: true, cellNF: true, WTF: true, UTC: true});
    assert(!(((wb||{}).Workbook||{}).WBProps||{}).date1904);
    var buf;
    {
      if(!wb.Workbook) wb.Workbook = {};
      if(!wb.Workbook.WBProps) wb.Workbook.WBProps = {};
      wb.Workbook.WBProps.date1904 = true;
      buf = XLSX.write(wb, {bookType: bookType, cellDates: true, type: "buffer", WTF: true});
      //require("fs").writeFileSync("1900_1904." + bookType, buf);
    }
    if(buf.l) delete buf.l;
    var wb2 = XLSX.read(buf, { cellDates: true, cellNF: true, WTF: true});
    assert((((wb2||{}).Workbook||{}).WBProps||{}).date1904);
    check_ws_dates(wb2.Sheets[wb2.SheetNames[0]], false, false, false, bookType == "xlsx");
  }); });
  if(typeof Uint8Array != "undefined") it("should cross 1900 -> 1904 Numbers", function() {
    var wb = XLSX.readFile("date/DateTime.xlsx", {cellDates: true, cellNF: true, WTF: true, UTC: true});
    assert(!(((wb||{}).Workbook||{}).WBProps||{}).date1904);
    var buf = XLSX.write(wb, { bookType: "numbers", cellDates: true, type: "buffer", WTF: true, numbers: numbers });
    //require("fs").writeFileSync("1900_1904.numbers", buf);
    if(buf.l) delete buf.l;
    var wb2 = XLSX.read(buf, { cellDates: true, cellNF: true, WTF: true});
    assert((((wb2||{}).Workbook||{}).WBProps||{}).date1904);
    check_ws_dates(wb2.Sheets[wb2.SheetNames[0]], false, false, false, false);
  });
  it("should cross 1904 -> 1900", function() {
  (["xlsx", "xlsb", "xls", "xlml", "biff5", "ods", "slk"] as Array<XLSX.BookType>).forEach(function(bookType) {
    var wb = XLSX.readFile("date1904/DateTime.xlsx", {cellDates: true, cellNF: true, WTF: true, UTC: true});
    assert((((wb||{}).Workbook||{}).WBProps||{}).date1904);
    delete wb.Workbook?.WBProps?.date1904;
    var buf = XLSX.write(wb, { bookType: bookType, cellDates: true, type: "buffer", WTF: true, numbers: numbers });
    //require("fs").writeFileSync("1904_1900.xls", buf);
    if(buf.l) delete buf.l;
    var wb2 = XLSX.read(buf, { cellDates: true, WTF: true});
    assert(!(((wb2||{}).Workbook||{}).WBProps||{}).date1904);
    check_ws_dates1904(wb2.Sheets[wb2.SheetNames[0]]);
  }); });
})
