const { readFile, read, write } = require("./"), assert = require("assert");
const numbers = require("./dist/xlsx.zahl");
const fs = require("fs");

/* BIFF2-4 and SYLK in Excel <= 5 encode \r instead of \n */
var replace_nl = [
  "excel5/I2960_21.xls",
  "excel5/I2960_3.xls",
  "excel5/I2960_4.xls",
  "excel3/I2960_32.xls",
  "excel3/I2960_33.xls",
  "excel2/I2960_21.xls",
  "excel2/I2960_20.xls",
  "excel2/I2960_22.xls",
  "excel5/i2960.slk",
];
/* BIFF5 in Excel5, SYLK in general does not support visibility */
var skip_vis = replace_nl.concat([
  "excel5/I2960_5.xls",
  "excel5/i2960.slk",
  "365/issue2960.biff5.xls",
  "365/issue2960.slk",
]);
var no_vis = [ "slk", "biff5", "biff4", "biff3", "biff2" ];
[
  ["excel5/I2960_21.xls", "biff2"],
  ["excel5/I2960_3.xls", "biff3"],
  ["excel5/I2960_4.xls", "biff4"],
  ["excel5/I2960_5.xls", "biff5"],
  ["excel3/I2960_32.xls", "biff2"],
  ["excel3/I2960_33.xls", "biff3"],
  ["excel2/I2960_20.xls", "biff2"],
  ["excel2/I2960_21.xls", "biff2"],
  ["excel2/I2960_22.xls", "biff2"],
  "excel5/i2960.slk",
  "365/issue2960.xlsx",
  "365/issue2960.xlsb",
  "365/issue2960.xls",
  "365/issue2960.xml",
  "365/issue2960.ods",
  "365/issue2960.slk",
  ["365/issue2960.biff5.xls", "biff5"],
].map(d => typeof d == "string" ? [d, d.slice(d.lastIndexOf(".")+1)]: d).forEach(([p, b]) => {
  /* R */
  console.log(p, b);
  const wb = readFile("i2960/" + p, { dense: false, WTF: 1 });
  check(wb, p, "");

  /* R -> W -> R */
  console.log(b);
  var out = write(wb, {bookType: b, type: "buffer", WTF:1, bookSST: b == "xls" ? true : false});
  if(b == "xlsb") fs.writeFileSync("blahblah.xlsb", out);
  var wb2 = read(out, {WTF:1});
  check(wb2, p, b);
});

[
  "tc/threaded.xlsx",
  "tc/threaded.xlsb",
  "tc/threaded.numbers",
].map(d => typeof d == "string" ? [d, d.slice(d.lastIndexOf(".")+1)]: d).forEach(([p, b]) => {
  /* R */
  console.log(p, b);
  const wb = readFile("i2960/" + p, { dense: false, WTF: 1 });
  checkT(wb, p);
  /* R -> W -> R */
  console.log(b);
  var out = write(wb, {bookType: b, type: "buffer", WTF:1, bookSST: b == "xls" ? true : false, numbers });
  if(b == "xlsb") fs.writeFileSync("blahblah.xlsb", out);
  if(b == "numbers") fs.writeFileSync("blahblah.numbers", out);
  var wb2 = read(out, {WTF:1});
  checkT(wb2, p, b);
});

function str(c,p) {
  var o = c.map(c=>c.t).join("");
  if(replace_nl.includes(p)) o = o.replace(/\r/g,"\n")
  return o;
}
function check(wb, p, b) {
  const ws = wb.Sheets[wb.SheetNames[0]];
  /* contents */
  assert.equal(str(ws["A1"].c, p), "Sheet JS:\nVisible");
  assert.equal(str(ws["A3"].c, p), "Sheet JS:\nHidden");
  /* visibility */
  if(skip_vis.includes(p)) return;
  if( b && no_vis.includes(b)) return;
  if(!b && no_vis.includes(p.slice(p.lastIndexOf(".")+1))) return;
  assert.equal(!!ws["A3"].c.hidden, true);
  assert.equal(!!ws["A1"].c.hidden, false);
}

function checkT(wb, p, b) {
  var parts = [
    "This is a test",
    "Is it really?",
    "Yes, 'tis",
    "Asd",
  ];
  const ws = wb.Sheets[wb.SheetNames[0]];
  const TC = ws["B2"].c;
  assert.equal(TC.length, 4);
  assert.equal(TC[0].a.replace(" ",""), "SheetJS");
  assert.equal(TC[0].t, parts[0]);
  for(var i = 1; i < 4; ++i) {
    assert.equal(TC[i].a, TC[0].a);
    assert.equal(TC[i].T, true);
    assert.equal(TC[i].t, parts[i]);
  }
}
